/*
 * $Id$
 */

#include <stdio.h>
#include <fstream>
#include <map>
#include <list>
#include <set>
#include <iomanip>
#include <iostream>
#include <cfloat>
#include <sstream>
#include "maxent.h"
#include "crf.h"
#include "common.h"

using namespace std;

string MODEL_DIR = "."; // the default directory for saving the models

const double PROB_OUTPUT_THRESHOLD = 0.01; // suppress output of tags with a very low probability

void
tokenize(const string & s1, vector<string> & vt);

void
ef_decode_beam(vector<Token> & vt, const vector<ME_Model> & vme,
	       vector< map<string, double> > & tagp);
void
crf_decode_forward_backward(Sentence & s, const CRF_Model & m, vector< map<string, double> > & tagp);

ParenConverter paren_converter;

void print_help()
{
  cout << "STEPP Tagger, a trainable part-of-speech tagger." << endl;
  cout << "Usage: stepp [OPTION]... [FILE]" << endl;
  cout << "Annotate each word in FILE with a part-of-speech tag." << endl;
  cout << "By default, the tagger assumes that FILE contains one sentence per line" << endl;
  cout << "and the words are tokenized with white spaces." << endl;
  //  cout << "Try -t and -s options if you want the tagger to process raw text." << endl;
  cout << "Try -t option if you want to process untokenized sentences." << endl;
  cout << endl;
  cout << "Mandatory arguments to long options are mandatory for short options too." << endl;
  cout << "  -m, --model=DIR        specify the directory containing the models" << endl;
  cout << "  -f, --fast             enable the fast tagging mode" << endl;
  cout << "  -t, --tokenize         perform tokenization" << endl;
  //  cout << "  -s, --split-sentence   perform sentence-splitting" << endl;
  cout << "  -p, --probability      output tag probabilities" << endl;
  cout << "  -h, --help             display this help and exit" << endl;
  cout << endl;
  cout << "Report bugs to <yoshimasa.tsuruoka@manchester.ac.uk>" << endl;
}

struct TagProb
{
  string tag;
  double prob;
  TagProb(const string & t_, const double & p_) : tag(t_), prob(p_) {}
  bool operator<(const TagProb & x) const { return prob > x.prob; }
};

int main(int argc, char** argv)
{
  bool PERFORM_TOKENIZATION = false;
  bool OUTPUT_TAG_PROBS = false;
  bool CRF_ONLY = false;
  
  string ifilename;

  for (int i = 1; i < argc; i++) {
    string v = argv[i];
    if ( (v == "-m" || v == "--model") && i < argc-1) {
      MODEL_DIR = argv[i+1];
      i++;
      continue;
    }
    if (v.substr(0, 8) == "--model=") {
      MODEL_DIR = v.substr(8);
      continue;
    }
    if (v == "-t" || v == "--tokenize") { 
      PERFORM_TOKENIZATION = true; 
      continue; 
    }
    if (v == "-p" || v == "--probability") { 
      OUTPUT_TAG_PROBS = true;
      continue; 
    }
    if (v == "-f" || v == "--fast") { 
      CRF_ONLY = true;
      continue; 
    }
    if (v == "-h" || v == "--help") {
      print_help();
      exit(0);
    }
    if (v[0] == '-') {
      cerr << "error: unknown option " << v << endl;
      cerr << "Try `stepp --help' for more information." << endl;
      exit(1);
    }
    ifilename = v;
  }

  istream *is(&std::cin);
  ifstream ifile;
  if (ifilename != "") {
    ifile.open(ifilename.c_str());
    if (!ifile) { cerr << "error: cannot open " << ifilename << endl; exit(1); }
    is = &ifile;
  }

  if (MODEL_DIR[MODEL_DIR.size()-1] != '/')  MODEL_DIR += "/";

  vector<ME_Model> vme(2);
  CRF_Model crfm;

  cerr << "loading the models from the directory \"" << MODEL_DIR << "\" ...";
  if (!crfm.load_from_file(MODEL_DIR + "model.crf")) exit(1);
  if (!CRF_ONLY) {
    if (!vme[0].load_from_file(MODEL_DIR + "model.ef.0")) exit(1);
    if (!vme[1].load_from_file(MODEL_DIR + "model.ef.1")) exit(1);
  }
  cerr << "done" << endl;

  //  crfm.save_to_file("test");
  
  string line;
  while (getline(*is, line)) {
    vector<string> vs;
    if (PERFORM_TOKENIZATION) {
      tokenize(line, vs);
    } else {
      istringstream is(line);
      string token;
      while (is >> token) vs.push_back(token);
    }

    // convert parantheses
    vector<Token> vt;
    vector<string> org_strs;
    for (vector<string>::const_iterator i = vs.begin(); i != vs.end(); i++) {
      string s = *i;
      org_strs.push_back(s);
      s = paren_converter.Ptb2Pos(s);
      vt.push_back(Token(s, "?"));
    }

    // tag the words
    vector< map<string, double> > tagp0, tagp1;
    crf_decode_forward_backward(vt, crfm, tagp0);
    if (!CRF_ONLY) {
      ef_decode_beam(vt, vme, tagp1);
    } else {
      for (vector<string>::const_iterator i = vs.begin(); i != vs.end(); i++) {
	map<string, double> dummy;
	tagp1.push_back(dummy);
      }
    }

    // merge the outputs (simple interpolation of probabilities)
    vector< map<string, double> > tagp; // merged
    for (size_t i = 0; i < vt.size(); i++) {
      const map<string, double> & crf = tagp0[i];
      const map<string, double> & ef  = tagp1[i];
      map<string, double> m, m2; // merged

      double sum = 0;
      for (map<string, double>::const_iterator j = crf.begin(); j != crf.end(); j++) {
	//	cout << j->first << ":" << j->second << " ";
	m.insert(pair<string, double>(j->first, j->second));
	sum += j->second;
      }
      //      cout << endl;

      for (map<string, double>::const_iterator j = ef.begin(); j != ef.end(); j++) {
	//	cout << j->first << ":" << j->second << " ";
	sum += j->second;
	if (m.find(j->first) == m.end()) {
	  m.insert(pair<string, double>(j->first, j->second));
	} else {
	  m[j->first] += j->second;
	}
      }
      //      cout << endl;

      const double th = PROB_OUTPUT_THRESHOLD * sum;
      for (map<string, double>::iterator j = m.begin(); j != m.end(); j++) {
	if (j->second >= th) m2.insert(*j);
      }
      double maxp = -1;
      string maxtag;
      for (map<string, double>::iterator j = m2.begin(); j != m2.end(); j++) {
	const double p = j->second;
	if (p > maxp) { maxp = p; maxtag = j->first; }
	//	cout << j->first << ":" << j->second << " ";
      }
      //      cout << endl;
      tagp.push_back(m2);
      vt[i].prd = maxtag;
    }

    // print the resutls
    for (size_t i = 0; i < vt.size(); i++) {
      const string s = org_strs[i];
      const string p = vt[i].prd;
      if (OUTPUT_TAG_PROBS) {
	vector<TagProb> tp;
	double sum = 0;
	for (map<string, double>::iterator j = tagp[i].begin(); j != tagp[i].end(); j++) {
	  tp.push_back(TagProb(j->first, j->second));
	  sum += j->second;
	}
	sort(tp.begin(), tp.end());
	cout << s;
	for (vector<TagProb>::iterator j = tp.begin(); j != tp.end(); j++) {
	  const double p = j->prob / sum; // normalize
	  if (p == 1) cout << resetiosflags(ios::fixed);
	  else        cout << setiosflags(ios::fixed) << setprecision(3);
	  cout << "\t" << j->tag << "\t" << p;
	}
	cout << endl;
      } else {
	if (i == 0) cout << s + "/" + p;
	else        cout << " " + s + "/" + p;
      }
    }
    cout << endl;
  }

}

/*
 * $Log$
 */

